<?php if (!defined('PmWiki')) exit();

/*
  suffr - administer a server democratically, via Wiki

  Copyright 2007,2008 Kirk Zurell <kirk@zurell.name>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  
  *** THIS SOFTWARE IS DESIGNED TO PERMIT EXECUTION OF ARBITRARY CODE
  *** ON COMPUTER SYSTEMS ON WHICH IT IS INSTALLED. IT ENJOYS ALL
  *** PRIVILEGES OF THE WEBSERVER PROCESS' userid AND group(s). IT IS
  *** INTENDED TO MANAGE FILES THAT DETERMINE THE SAFE AND CONSISTENT
  *** OPERATION OF THE SERVER. IT DOES NOT PREVENT MISUSE OF THESE
  *** FILES OR OF THE SYSTEM ITSELF.

*/

/** \file suffr.php
 * \brief manage a wiki and server democratically
 *
 * See http://www.pmwiki.org/wiki/Cookbook/Suffr
 *
 * This is a module for PMWiki, a PHP wiki found at www.pmwiki.org. It
 * allows for management of the configuration files of the
 * wiki/website's host system, according to a democratic regime.
 *
 * For more documentation, see:
 *  - the page PmWiki.Suffr
 *  - the source code
 *  - the README
 *
 * See the important information in the suffr.php source file that
 * describes intended effects and hazards of using this software.
 */

$RecipeInfo['suffr']['Version'] = '2008-01-16';


/* CONFIGURATION
   Change these values in your local/config.php if desired.

   These values need to match the $WikiDir created with the PageStore
   class. suffr looks through this directory for Profiles pages.

   For instance, to change $WikiDir and these to use wiki.d/Profiles:
   in config.php:
     $WikiDir = new PageStore('wiki.d/$Group/$FullName');
     //...
     $suffrProfilesDir = "wiki.d/Profiles";
*/

//suffr looks here for Profiles pages
SDV($suffrProfilesDir,"wiki.d");
SDV($suffrPagePattern,"/(($GroupPattern)([\\/.])?($NamePattern))/");
SDV($suffrProfilesGroup, "Profiles");
SDV($suffrProfilesGroupHome, "Profiles");

//suffr approves pages when this percentage of users sign off on a
//page
SDV($suffrSuccessThreshold, 50/100);



/* SUFFR module programming */


/* Allow users to sign a page */

Markup("suffrpagesignature", 
       'directives', 
       '/\\(:pagesignature:\\)/e',
       "Keep(suffrGenerateSignature());");

//should perform author check.
$suffrPageSignFmt = "<a href='{\$PageUrl}?action=sign'>\$LinkText</a>";
function suffrGenerateSignature()
{
  global $Author,$AuthorGroup,$pagename,$suffrPageSignFmt;
  return MakeLink($pagename,//should be user's profile page?
		  $pagename,
		  'Sign this page',
		  '',
		  $suffrPageSignFmt);
}



/* Sign a page, and check whether page actions should take place */

$HandleActions['sign'] = 'suffrSignPage';

function suffrSignPage($pagename, $auth='read')
{
  global $Author,$AuthorPage;
  global $GroupPattern,$NamePattern;
  global $PageStartFmt, $PageEndFmt;
  global $suffrProfilesDir;
  global $suffrSuccessThreshold;
  global $suffrPagePattern;
  global $suffrProfilesGroup;
  global $suffrProfilesGroupHome;

  $signatures = 0; $total = 0;
  $SignResultFmt = "";


  /* calculate md5sum of signed page */
  $orig = ReadPage($pagename);//release?
  $md5sum = md5($orig['text']);
  $action = "";

  /* if we know the author, sign this page on the AuthorPage */
  if(isset($Author))
    {
      $signcount = 0;
      $page = ReadPage($AuthorPage);
      $new = $page;
      //replace existing invalid signature or add new signature
      $new['text'] = preg_replace("/\n?\(:signature (($GroupPattern([\\/.]))?($NamePattern)) [0-9a-f]+ :\)|\$/",
				  "\n(:signature $pagename $md5sum :)",
				  $new['text'],
				  1,
				  $signcount);
      $SignResultFmt .= "<p>Signatures added/replaced: $signcount</p>";

      //$new['text'] .= "\n(:signature $pagename $md5sum :)";

      //post updated page text.
      $_POST['post'] = 'y';
      PostPage($AuthorPage, $page, $new);
    }
  else
    {
      //we don't know the author.
      //got here in error
      Abort("Problem: must know author");
    }


  /* check all Profiles pages (except RecentChanges and home page) */

  $pdir = @opendir($suffrProfilesDir);
  if ($pdir) 
    { 
      while ( ($each = @readdir($pdir)) !== false) 
	{
	  $matches = Array(); 
	  preg_match($suffrPagePattern,$each,$matches);

	  //eliminate all pages not actual profile pages
	  if(@$matches[2] !== $suffrProfilesGroup) continue;
	  if(@$matches[4] === "RecentChanges") continue;
	  if(@$matches[4] === $suffrProfilesGroupHome) continue;
	  if(@$matches[4] === "HomePage") continue;

	  //
	  $total++;

	  if(/*matches user*/ @$AuthorPage == $matches[2] . '/' . $matches[4])
	    {
	      //already signed
	      $signatures++;
	    }
	  else
	    {
	      $page = ReadPage($each);
	      $signed = @preg_match("/\\(:signature $pagename $md5sum :\\)/",
				   $page['text']);			 
	      if($signed > 0) $signatures++;
	    }
	}
    }
  @closedir($pdir);

  //debug
  $SignResultFmt .= "<p>sigs $signatures / total $total</p>";

  /* If page is signed by a majority, perform action */

  $OldLogPattern = "/\\n\(:log \s+ (.*?) \s* :\)" //NASTY HACK:
						  //removes return
						  //attached to log.
    . "(.*?)"
    . "\(:log \s* :\)/sx";
  $CmdPattern = '/\(:command \s+ (.+?) \s* :\)/ex';
  $RCFilePattern = "/\(:rcfile \s+ (.+?) \s* :\)" //option matches
						  //newlines, must fix
    . "(.*?)"
    . "\(:rcfile \s* :\)/sex";
  $ScriptPattern = "/\(:script \s* :\)" //option matches newlines,
					//must fix
    . "(.*?)"
    . "\(:script \s* :\)/sex";



  if($total > 0)
    if(($signatures/$total) > $suffrSuccessThreshold)
      {
	//page is authorized to execute.
	$count = 0;


	$orig['text'] = preg_replace(
			    Array(
				  $OldLogPattern,
				  $CmdPattern,
				  $RCFilePattern,
				  $ScriptPattern
				  ),
			    Array(
				  '', //empty to remove old log entries.
				  "PSS('$0') . suffrInvokeCommand('$0','$1')",
				  "PSS('$0') . suffrCopyRCFile('$0','$1','$2')",
				  "PSS('$0') . suffrInvokeScript('$0','$1')"
				  ),
			    $orig['text'],
			    -1, //no limit
			    $count);

	$SignResultFmt .= "\nActions carried out ($count).\n\nSee [[$pagename]]\n";

	

	//write results back to page
	WritePage($pagename,$orig);

	//notify everyone?

      }


  /* Finish; show results of signature */
  SDV($suffrPageSignFmt, "<p>Page is signed</p>$SignResultFmt");
  SDV($suffrHandleSignFmt,Array(&$PageStartFmt, &$suffrPageSignFmt, &$PageEndFmt));
  PrintFmt($pagename,$suffrHandleSignFmt);

}

function suffrPMLog($message = 'Error')
{
  $logtext = "\n(:log " . date(DATE_RFC822) . " :)\n";
  $logtext .= $message . "\n";
  
  if($EnableDiag and error_get_last() !== NULL)
     $logtext .=  print_r(error_get_last(),true);

  $logtext .= "(:log:)";
  return $logtext;
}

//commands are visible to ps
function suffrInvokeCommand($whole,$command)
{
  $outlines = Array();
  $retval = '';

  $last = exec($command,$outlines,$retval);

  if($retval != 0)
    {
      return suffrPMLog("Command failed ($retval):\n" . implode("\n",$outlines));
    }
  else
    {
      return suffrPMLog("Command succeeded ($retval):\n" . implode("\n",$outlines));
    }

}

function suffrCopyRCFile($whole,$filename,$body)
{

  if($filename == '') return;
  
  //check if file is writable
  //if(!is_writable(dirname($filename)))
  //  {
  //   return suffrPMLog('unable to write');
  //  }
  //create enclosing directory
  $exists = is_dir(dirname($filename));
  if(!$exists)
    if(!@mkdir(dirname($filename),0777,true)) //recursive
      {
	return suffrPMLog('unable to make directory');
      }
  
  //open file
  if(!($handle = @fopen($filename,'w')))
    {
      return suffrPMLog('unable to open');
    };

  //flock($handle,LOCK_EX);flock($handle,LOCK_UN);

  //write contents
  if(!fwrite($handle,$body))
    $log = "File not written";
  else
    $log = "Copied file";


  return suffrPMLog($log);
}

function suffrInvokeScript($whole,$script)
{
  $filename = tempnam(sys_get_temp_dir(),'wiki');

  if(!($handle = @fopen($filename,'w')))
    {
      return suffrPMLog('unable to create script');
    }

  if(!(fwrite($handle,"#!/bin/bash\n") && fwrite($handle,$script)))
    {
      return suffrPMLog('unable to write script');
    }

  fclose($handle);

  if(!chmod($filename,0700))
    return suffrPMLog("unable to chmod");

  $outlines = Array();
  $last = exec($filename,$outlines,$retval);

  if($retval != 0)
    {
      return suffrPMLog("Script failed ($retval):\n" . implode("\n",$outlines));
    }
  else
    {
      return suffrPMLog("Script succeeded ($retval):\n" . implode("\n",$outlines));
    }
  
}


/* Display signatures, and whether valid or not */


SDV($suffrPageSignValidFmt, "<a href=\"\$LinkUrl\">\$LinkText</a>");
SDV($suffrMD5Fmt, "[0-9a-f]+");

Markup("suffrsignature",
       'directives',
       "/\\(:signature (($GroupPattern([\\/.]))?($NamePattern)) ($suffrMD5Fmt) :\\)/e",
       "'<br/>' . Keep(suffrCheckSignature('$1','$5'))");

function suffrCheckSignature($target, $md5sum)
{
  global $pagename;
  
  $page = ReadPage($target);
  $totest = md5($page['text']);

  $valid = ($totest === $md5sum);

  return MakeLink($pagename,
		  PSS($target),
		  NULL,//use target name
		  ($valid ? " (valid)" : " (invalid)"),
		  $suffrPageSignValidFmt,
		  'L');
}



/* Markup/Display of sysadmin elements. */

Markup('suffrcommand','inline',
       "/\\(:command (.+) :\\)/e",
       "Keep(suffrDisplayCommand(\$pagename,PSS('$1')))");
$HTMLStylesFmt['command'] = 'p.command { background-color: #FFFFDD; }';

function suffrDisplayCommand($pagename,$command)
{
  return "\n<p class='command'><em>Command:</em> $command</p>";
}

Markup('suffrrcfile', 'fulltext',
       "/\\(:rcfile \\s+ ([^\n]+?) \\s* :\\)
         (.*?)
         \\(:rcfile \\s* :\\)/seix",
       "Keep(suffrDisplayRCFile(\$pagename, PSS('$1'), PSS('$2')))");
$HTMLStylesFmt['rcfile-legend'] = 'p.legend { background-color: #EEFFFF; }';
$HTMLStylesFmt['rcfile'] = 'pre.rcfile { background-color: #DDFFFF; }';

function suffrDisplayRCFile($pagename, $filename, $body)
{
  $output = "\n<p class='legend'><em>File:</em> $filename</p>\n";
  $output .= "<pre class='rcfile'>";
  $output .= $body;
  $output .= "</pre>\n";
  return $output;
}

Markup('suffrscript', 'fulltext',
       "/\\(:script \\s* :\\)
         (.*?)
         \\(:script \\s* :\\)/seix",
       "Keep(suffrDisplayScript(\$pagename, PSS('$1')))");
$HTMLStylesFmt['script-legend'] = 'p.script { background-color: #EEEEFF; }';
$HTMLStylesFmt['script'] = 'pre.script { background-color: #DDDDFF; }';

function suffrDisplayScript($pagename, $body)
{
  $output = "\n<p class='script'><em>Script:</em> #!/bin/bash</p>\n";
  $output .= "<pre class='script'>";
  $output .= $body;
  $output .= "</pre>\n";
  return $output;
}

Markup('suffrlog','fulltext',
       "/\\(:log \\s+ (.*?) \\s* :\\)
        (.*?)
        \\(:log \\s* :\\)/seix",
       "Keep(suffrDisplayLog(\$pagename, '$1', PSS('$2')))");
$HTMLStylesFmt['log-legend'] = 'p.log-legend { background-color: #FFBBFF; }';
$HTMLStylesFmt['log'] = 'pre.log { background-color: #FFDDFF; }';

function suffrDisplayLog($pagename, $logdate, $logtext)
{
  $output = "\n<p class='log-legend'><em>Log:</em> $logdate</p>";
  $output .= "\n<pre class='log'>";
  $output .= $logtext;
  $output .= "</pre>\n";
  return $output;
}




/* Outstanding issues:
 * How to avoid multiple re-signings: edit the page w/ result to invalidate?
*/


/*
 * For redundancy and greater certainty, the author(s) _repeat_ that
 * they offer no warranty as to the safety or effectiveness of this
 * software.
 */
